#!/bin/bash
#For each line in TestInputs.txt file, execute simulator based on line's contents
#And compare output log file to various expected outputs in line's contents.
PRINT_FAILS_ONLY=1
RC=0 # Return Code (0 == success; 1 == failure)
if [ -n $2 ]
then
	if [ "-all" == "$2" ]
	then
		PRINT_FAILS_ONLY=0
	fi
fi
TOLERANCE=0

if [ -z $1 ] || [ ! -f $1 ]
then
	echo "Invalid input file."
	RC=1
	exit $RC
fi

#Arguments: (trialNum, logLocation, expectedVal)
emptyLog (){
	local NUMLINES=0
	while read -r OUTPUT_LINE; do
		case ${OUTPUT_LINE} in
			[0-9]*)
			((NUMLINES++))
		esac
	done < $1

	if [ $NUMLINES -eq 0 ]
	then
		echo 1
	else
		echo 0
	fi
}

numGC () {
	local NUMLINES=0
	while read -r OUTPUT_LINE; do
		case ${OUTPUT_LINE} in
			[0-9]*)
			((NUMLINES++))
		esac
	done < $2

	local returnVal=$(compare $3 $NUMLINES)
	if [ $returnVal -eq -1 ]
	then
		echo "$1 failed: Number of GC's ($NUMLINES) too different from expected value ($3)"
		RC=1
	else
		if [ $PRINT_FAILS_ONLY -eq 0 ]
		then
			echo "$1 success: Number of GC's"
		fi
	fi
}

memUsed (){
	local finalFree
	finalFree=$(removeNonDataLines $2 | sed -r 's/\s+//g' | awk -F'\\|' '{} END{print $6}')

	local returnVal=$(compare $3 $finalFree)
	if [ $returnVal -eq -1 ]
	then
		echo "$1 failed: Final heap size ($finalFree) too different from expected value ($3)"
		RC=1
	else
		if [ $PRINT_FAILS_ONLY -eq 0 ]
		then
			echo "$1 success: Final heap size"
		fi
	fi
}

lastCollectionReason(){
	local reason
	reason=$(removeNonDataLines $2 | sed -r 's/\s+//g' | awk -F'\\|' '{} END{print $2}')
	if [ $reason == $3 ]
	then
		if [ $PRINT_FAILS_ONLY -eq 0 ]
		then
			echo "$1 success: Last collection reason"
		fi
	else
		echo "$1 failed: Last collection reason ($reason) does not match expected reason ($3)"
		RC=1
	fi
}

simCrashed(){
	local lastLine
	lastLine=$(tail -1 $2 | sed -r 's/\s+//g')
	if [[ $lastLine =~ ^[0-9].* ]]
	then
		echo "$1 failed: Simulator crashed during execution."
		RC=1
	fi
}

#Arguments: (Expected value, Experimental value)
compare(){
	local expectedVal=$1
	local experimentalVal=$2
	local tol=$(echo "$expectedVal * $TOLERANCE" | bc -l)
	local lowerBound=$(echo "$expectedVal-$tol" | bc -l)
	local upperBound=$(echo "$expectedVal+$tol" | bc -l)

	local lowerCompare=$(echo "$experimentalVal < $lowerBound" | bc -l)
	local upperCompare=$(echo "$experimentalVal > $upperBound" | bc -l)
	if [ $lowerCompare = 1 ] || [ $upperCompare = 1 ]
	then
		echo -1
	else
		echo 1
	fi
}

function removeNonDataLines() {
	sed '/^\W*[1-9]/!d' $1
}


#Elements of LINE: (trialNum, commandLine, logLocation, [various tests])
while IFS=';' read -ra LINE; do
	case "${LINE[0]}" in
		\#*) continue ;;
		"") continue ;;
	esac

#Run simulator
	uniqueID=$(($(date +%s%N)/1000000))
	simCall="${LINE[1]} -li $uniqueID"
	eval $simCall &> /dev/null

#Check if log file exists, and if unique identifier on first line matches
	if [ ! -e ${LINE[2]} ]
	then
		echo "${LINE[0]} failed: Log file not found."
		RC=1
		continue;
	fi
	
	logEmpty=$(emptyLog ${LINE[2]})
	if [ $logEmpty -eq 1 ]
	then
		echo "${LINE[0]} failed: Log file is empty. Simulator may have crashed early in execution."
		RC=1
		continue
	fi

	idLine=$(head -n 1 ${LINE[2]})
	logID=$(echo $idLine | awk '{print $2}')
	if [ $logID -ne $uniqueID ]
	then
		echo "${LINE[0]} failed: Log file exists, but was not created by this run of the simulator."
		RC=1
		continue
	fi

#NEW TEST CALLS SHOULD GO HERE
	numGC "${LINE[0]}" ${LINE[2]} ${LINE[3]}
	memUsed "${LINE[0]}" ${LINE[2]} ${LINE[4]}
	lastCollectionReason "${LINE[0]}" ${LINE[2]} ${LINE[5]}
	simCrashed "${LINE[0]}" ${LINE[2]}
done < $1

if [ $RC -ne 0 ]
then
	echo -e "\nERROR(s) were encountered! RC: $RC\n" >&2
	echo -e "Please Verify your changes and the Regression tests.\n" >&2
	echo -e "You MUST CORRECT one or both before you are able to push!\n" >&2
	echo -e "Please see ../../CONTRIBUTING.md\n" >&2
fi

exit $RC
